-module(tb_server).
-export([start/2, stop/0]).
-compile([debug_info]).

start(Port, Opts) ->
    spawn(fun() -> start_server(Port, Opts) end).

stop() ->
    case whereis(tb_server) of
        undefined ->
            ok;
        Pid ->
            exit(Pid, kill)
    end.

start_server(Port, Opts) ->
    process_flag(trap_exit, true),
    TcpOpts = make_tcp_opts(Opts),
    case gen_tcp:listen(Port, [binary|TcpOpts]) of
        {ok, LSock} ->
            Pid = spawn_link(fun() -> listener(LSock, TcpOpts) end),
            register(tb_server, Pid),
            gen_tcp:controlling_process(LSock, Pid);
        Error ->
            io:format("Listen error: ~p\n", [Error]),
            erlang:error(Error)
    end.

listener(LSock, Opts) ->
    process_flag(trap_exit, true),
    case gen_tcp:accept(LSock) of
        {ok, Sock} ->
            Pid = spawn(fun() -> loop(Sock, Opts) end),
            gen_tcp:controlling_process(Sock, Pid),
            listener(LSock, Opts);
        {error, timeout} ->
            io:format("Socket timed out\n");
        Error ->
            io:format("Accept error: ~p\n", [Error])
    end,
    (catch unregister(tb_server)).

loop(Sock, Opts) ->
    StartTime = now(),
    case loop(Sock, Opts, 0, StartTime) of
        {ok, {stats, _}} ->
            ok;
        Error ->
            io:format("Error in loop/2: ~p\n", [Error])
    end.

loop(Sock, Opts, BytesRecv, StartTime) ->
    receive
        {tcp, Sock, Data} ->
            inet:setopts(Sock, Opts),
            loop(Sock, Opts, BytesRecv + byte_size(Data), StartTime);
        {tcp_closed, Sock} ->
            EndTime = now(),
            io:format("Socket closed\n"),
            Stats = tb_util:stats(StartTime, EndTime, BytesRecv),
            io:format("~s", [tb_util:format_stats(Stats)]),
            {ok, Stats};
        {tcp_error, Sock, Reason} ->
            io:format("tcp_error: ~p\n", [Reason]),
            gen_tcp:close(Sock),
            {error, Reason}
    end.

make_tcp_opts(Opts) when is_list(Opts) ->
    SrtDefOpts = lists:keysort(1, [
        {active, once},
        {reuseaddr, true},
        {packet, 0},
        {packet_size, 65536},
        {recbuf, 1000000}
    ]),
    SrtOpts = lists:keysort(1, Opts),
    lists:ukeymerge(1, SrtOpts, SrtDefOpts).

